@@ -2,7 +2,7 @@
 
  dxf_parser.c -- implements DXF support [parsing]
 
- version 4.2, 2014 July 25
+ version 5.0, 2020 August 1
 
  Author: Sandro Furieri a.furieri@lqt.it
 
@@ -24,7 +24,7 @@ The Original Code is the SpatiaLite library
 
 The Initial Developer of the Original Code is Alessandro Furieri
  
-Portions created by the Initial Developer are Copyright (C) 2008-2013
+Portions created by the Initial Developer are Copyright (C) 2008-2020
 the Initial Developer. All Rights Reserved.
 
 Contributor(s): 
@@ -43,6 +43,15 @@ the terms of any one of the MPL, the GPL or the LGPL.
  
 */
 
+/*
+ 
+CREDITS:
+
+inital development of the DXF module has been funded by:
+Regione Toscana - Settore Sistema Informativo Territoriale ed Ambientale
+
+*/
+
 #include <stdlib.h>
 #include <stdio.h>
 #include <string.h>
@@ -1299,7 +1308,7 @@ static void
 unlinked_rings (const void *p_cache, gaiaDxfPolylinePtr line)
 {
 /* attempt to identify unlinked Polygon rings */
-    int invalid;
+    int invalid = 0;
     int start;
     int count;
     double x;
@@ -2412,9 +2421,15 @@ reset_dxf_entity (const void *p_cache, gaiaDxfParserPtr dxf)
 	  dxf->curr_layer_name = NULL;
       }
     if (dxf->extra_key != NULL)
-	free (dxf->extra_key);
+      {
+	  free (dxf->extra_key);
+	  dxf->extra_key = NULL;
+      }
     if (dxf->extra_value != NULL)
-	free (dxf->extra_value);
+      {
+	  free (dxf->extra_value);
+	  dxf->extra_value = NULL;
+      }
     ext = dxf->first_ext;
     while (ext != NULL)
       {
@@ -3522,7 +3537,7 @@ gaiaParseDxfFileCommon (const void *p_cache, gaiaDxfParserPtr dxf,
 		p = line;
 		continue;
 	    }
-	  *p++ = c;
+	  *p++ = (char) c;
 	  /* Even Rouault 2013-06-02 - avoiding a potential buffer overflow */
 	  if (p - line == sizeof (line) - 1)
 	      goto stop;
