@@ -2,7 +2,7 @@
 
  gg_sqlaux.c -- SQL ancillary functions
 
- version 4.2, 2014 July 25
+ version 5.0, 2020 August 1
 
  Author: Sandro Furieri a.furieri@lqt.it
 
@@ -24,7 +24,7 @@ The Original Code is the SpatiaLite library
 
 The Initial Developer of the Original Code is Alessandro Furieri
  
-Portions created by the Initial Developer are Copyright (C) 2008-2013
+Portions created by the Initial Developer are Copyright (C) 2008-2020
 the Initial Developer. All Rights Reserved.
 
 Contributor(s):
@@ -48,6 +48,7 @@ the terms of any one of the MPL, the GPL or the LGPL.
 #include <stdio.h>
 #include <string.h>
 #include <math.h>
+#include <ctype.h>
 
 #if defined(_WIN32) && !defined(__MINGW32__)
 #include "config-msvc.h"
@@ -1147,6 +1148,13 @@ gaiaParseDMS (const char *dms, double *longitude, double *latitude)
 GAIAAUX_DECLARE char *
 gaiaConvertToDMS (double longitude, double latitude)
 {
+/* formatting a DMS string */
+    return gaiaConvertToDMS (longitude, latitude);
+}
+
+GAIAAUX_DECLARE char *
+gaiaConvertToDMSex (double longitude, double latitude, int decimal_digits)
+{
 /* formatting a DMS string */
     char *dms0;
     char *dms;
@@ -1155,11 +1163,17 @@ gaiaConvertToDMS (double longitude, double latitude)
     int long_d;
     int long_m;
     int long_s;
+    double long_s_dbl;
     int lat_d;
     int lat_m;
     int lat_s;
+    double lat_s_dbl;
     double val;
     int len;
+    if (decimal_digits < 0)
+	decimal_digits = 0;
+    if (decimal_digits > 8)
+	decimal_digits = 8;
     if (longitude < -180.0 || longitude > 180.0)
 	return NULL;
     if (latitude < -90.0 || latitude > 90.0)
@@ -1174,27 +1188,324 @@ gaiaConvertToDMS (double longitude, double latitude)
 	  lat_prefix = 'S';
 	  latitude *= -1.0;
       }
-    long_d = floor (longitude);
+    long_d = (int) floor (longitude);
     val = 60.0 * (longitude - (double) long_d);
-    long_m = floor (val);
+    long_m = (int) floor (val);
     val = 60.0 * (val - (double) long_m);
-    long_s = floor (val);
+    long_s_dbl = val;
+    long_s = (int) floor (val);
     if ((val - (double) long_s) > 0.5)
 	long_s++;
-    lat_d = floor (latitude);
+    lat_d = (int) floor (latitude);
     val = 60.0 * (latitude - (double) lat_d);
-    lat_m = floor (val);
+    lat_m = (int) floor (val);
     val = 60.0 * (val - (double) lat_m);
-    lat_s = floor (val);
+    lat_s = (int) floor (val);
+    lat_s_dbl = val;
     if ((val - (double) lat_s) > 0.5)
 	lat_s++;
-    dms0 =
-	sqlite3_mprintf ("%02d°%02d′%02d″%c %03d°%02d′%02d″%c", lat_d,
-			 lat_m, lat_s, lat_prefix, long_d, long_m, long_s,
-			 long_prefix);
+    if (decimal_digits == 0)
+	dms0 =
+	    sqlite3_mprintf ("%02d°%02d′%02d″%c %03d°%02d′%02d″%c",
+			     lat_d, lat_m, lat_s, lat_prefix, long_d, long_m,
+			     long_s, long_prefix);
+    else
+      {
+	  char format[256];
+	  sprintf (format,
+		   "%%02d°%%02d′%%0%d.%df″%%c %%03d°%%02d′%%0%d.%df″%%c",
+		   decimal_digits + 3, decimal_digits, decimal_digits + 3,
+		   decimal_digits);
+	  dms0 =
+	      sqlite3_mprintf (format, lat_d, lat_m, lat_s_dbl, lat_prefix,
+			       long_d, long_m, long_s_dbl, long_prefix);
+      }
     len = strlen (dms0);
     dms = malloc (len + 1);
     strcpy (dms, dms0);
     sqlite3_free (dms0);
     return dms;
 }
+
+#if OMIT_ICONV == 0		/* ICONV is absolutely required */
+
+/*********************************************************************
+/
+/ DISCLAIMER
+/
+/ the following code implementation (URL percent-encoding/-decoding)
+/ simply is a rearranged adaption of this original code released
+/ into the Public Domain:
+/
+/ http://www.geekhideout.com/urlcode.shtml
+/
+*********************************************************************/
+
+static char
+url_to_hex (char code)
+{
+    static char hex[] = "0123456789abcdef";
+    return hex[code & 15];
+}
+
+GAIAAUX_DECLARE char *
+gaiaEncodeURL (const char *url, const char *out_charset)
+{
+/* encoding some URL */
+    char *encoded = NULL;
+    const char *in;
+    char *utf8_url = NULL;
+    char *out;
+    int len;
+    if (url == NULL)
+	return NULL;
+
+    utf8_url = url_fromUtf8 (url, out_charset);
+    if (utf8_url == NULL)
+	return NULL;
+    len = strlen (url);
+    if (len == 0)
+	return NULL;
+    in = utf8_url;
+
+    encoded = malloc ((len * 3) + 1);
+    out = encoded;
+    while (*in != '\0')
+      {
+	  if (isalnum (*in) || *in == '-' || *in == '_' || *in == '.'
+	      || *in == '~')
+	      *out++ = *in;
+	  else
+	    {
+		*out++ = '%';
+		*out++ = url_to_hex (*in >> 4);
+		*out++ = url_to_hex (*in & 15);
+	    }
+	  in++;
+      }
+    *out = '\0';
+    free (utf8_url);
+    return encoded;
+}
+
+static char
+url_from_hex (char ch)
+{
+    return isdigit (ch) ? ch - '0' : (char) (tolower (ch) - 'a' + 10);
+}
+
+GAIAAUX_DECLARE char *
+gaiaDecodeURL (const char *encoded, const char *in_charset)
+{
+/* decoding some URL */
+    char *url = NULL;
+    char *utf8_url = NULL;
+    const char *in = encoded;
+    char *out;
+    int len;
+    if (encoded == NULL)
+	return NULL;
+    len = strlen (encoded);
+    if (len == 0)
+	return NULL;
+
+    url = malloc (len + 1);
+    out = url;
+    while (*in != '\0')
+      {
+	  if (*in == '%')
+	    {
+		if (*(in + 1) && *(in + 2))
+		  {
+		      *out++ =
+			  url_from_hex (*(in + 1)) << 4 |
+			  url_from_hex (*(in + 2));
+		      in += 2;
+		  }
+	    }
+	  else if (*in == '+')
+	      *out++ = ' ';
+	  else
+	      *out++ = *in;
+	  in++;
+      }
+    *out = '\0';
+    utf8_url = url_toUtf8 (url, in_charset);
+    free (url);
+    return utf8_url;
+}
+
+#endif /* ICONV enabled/disabled */
+
+GAIAAUX_DECLARE char *
+gaiaDirNameFromPath (const char *path)
+{
+/* extracting the DirName (if any) from a Path */
+    const char *in = path;
+    const char *last = NULL;
+    int len = 0;
+    int dirlen = len;
+    char *name;
+
+    if (path == NULL)
+	return NULL;
+
+    while (*in != '\0')
+      {
+	  /* parsing the Path */
+	  len++;
+	  if (*in == '/' || *in == '\\')
+	    {
+		last = in;
+		dirlen = len;
+	    }
+	  in++;
+      }
+    if (last == NULL)
+	return NULL;		/* there is no Dir component */
+
+/* allocating the DirName  to be returned */
+    name = malloc (dirlen + 1);
+    memcpy (name, path, dirlen);
+    *(name + dirlen) = '\0';
+    return name;
+}
+
+GAIAAUX_DECLARE char *
+gaiaFullFileNameFromPath (const char *path)
+{
+/* extracting the FullFileName (including Extension) from a Path */
+    const char *in = path;
+    const char *last = path - 1;
+    int len;
+    char *name;
+
+    if (path == NULL)
+	return NULL;
+
+    while (*in != '\0')
+      {
+	  /* parsing the Path */
+	  if (*in == '/' || *in == '\\')
+	      last = in;
+	  in++;
+      }
+    len = strlen (last + 1);
+    if (len == 0)
+	return NULL;
+
+/* allocating the FullFileName to be returned */
+    name = malloc (len + 1);
+    strcpy (name, last + 1);
+    return name;
+}
+
+GAIAAUX_DECLARE char *
+gaiaFileNameFromPath (const char *path)
+{
+/* extracting the FileName (excluding Extension) from a Path */
+    const char *in = path;
+    const char *last = path - 1;
+    int len;
+    char *name;
+    int i;
+
+    if (path == NULL)
+	return NULL;
+
+    while (*in != '\0')
+      {
+	  /* parsing the Path */
+	  if (*in == '/' || *in == '\\')
+	      last = in;
+	  in++;
+      }
+    len = strlen (last + 1);
+    if (len == 0)
+	return NULL;
+
+/* allocating the FullFileName to be returned */
+    name = malloc (len + 1);
+    strcpy (name, last + 1);
+    for (i = len - 1; i > 0; i--)
+      {
+	  if (*(name + i) == '.')
+	    {
+		/* stripping out the extension */
+		*(name + i) = '\0';
+		break;
+	    }
+      }
+    return name;
+}
+
+GAIAAUX_DECLARE char *
+gaiaFileExtFromPath (const char *path)
+{
+/* extracting the FileExtension (if any) from a Path */
+    int len;
+    char *name;
+    int i;
+    int pos = -1;
+
+    if (path == NULL)
+	return NULL;
+
+    len = strlen (path);
+    for (i = len - 1; i > 0; i--)
+      {
+	  if (*(path + i) == '/' || *(path + i) == '\\')
+	      break;
+	  if (*(path + i) == '.')
+	    {
+		/* found an extension */
+		pos = i;
+		break;
+	    }
+      }
+    if (pos <= 0)
+	return NULL;
+
+/* allocating the FileExtension to be returned */
+    len = strlen (path + pos + 1);
+    if (len == 0)
+	return NULL;
+    name = malloc (len + 1);
+    strcpy (name, path + pos + 1);
+    return name;
+}
+
+GAIAAUX_DECLARE char *
+gaiaRemoveExtraSpaces (const char *string)
+{
+/* removing all repeated whitespaces from a string */
+    int len;
+    char *clean;
+    char *p;
+    int i;
+    int space = 0;
+
+    if (string == NULL)
+	return NULL;
+
+    len = strlen (string);
+    clean = malloc (len + 1);
+    p = clean;
+    for (i = 0; i < len; i++)
+      {
+	  int sp = 0;
+	  if (string[i] == ' ' || string[i] == '\t')
+	      sp = 1;
+	  else
+	      sp = 0;
+	  if (space && sp)
+	      continue;
+	  *p++ = string[i];
+	  if (sp)
+	      space = 1;
+	  else
+	      space = 0;
+      }
+    *p = '\0';
+    return clean;
+}
