@@ -2,7 +2,7 @@
 
  virtualdbf.c -- SQLite3 extension [VIRTUAL TABLE accessing DBF]
 
- version 4.2, 2014 July 25
+ version 5.0, 2020 August 1
 
  Author: Sandro Furieri a.furieri@lqt.it
 
@@ -24,7 +24,7 @@ The Original Code is the SpatiaLite library
 
 The Initial Developer of the Original Code is Alessandro Furieri
  
-Portions created by the Initial Developer are Copyright (C) 2008-2013
+Portions created by the Initial Developer are Copyright (C) 2008-2020
 the Initial Developer. All Rights Reserved.
 
 Contributor(s):
@@ -77,6 +77,7 @@ typedef struct VirtualDbfStruct
     char *zErrMsg;		/* error message: USE INTERNALLY BY SQLITE */
     sqlite3 *db;		/* the sqlite db holding the virtual table */
     gaiaDbfPtr dbf;		/* the DBF struct */
+    int text_dates;
 } VirtualDbf;
 typedef VirtualDbf *VirtualDbfPtr;
 
@@ -105,6 +106,31 @@ typedef struct VirtualDbfCursorStruct
 
 typedef VirtualDbfCursor *VirtualDbfCursorPtr;
 
+static char *
+convert_dbf_colname_case (const char *buf, int colname_case)
+{
+/* converts a DBF column-name to Lower- or Upper-case */
+    int len = strlen (buf);
+    char *clean = malloc (len + 1);
+    char *p = clean;
+    strcpy (clean, buf);
+    while (*p != '\0')
+      {
+	  if (colname_case == GAIA_DBF_COLNAME_LOWERCASE)
+	    {
+		if (*p >= 'A' && *p <= 'Z')
+		    *p = *p - 'A' + 'a';
+	    }
+	  if (colname_case == GAIA_DBF_COLNAME_UPPERCASE)
+	    {
+		if (*p >= 'a' && *p <= 'z')
+		    *p = *p - 'a' + 'A';
+	    }
+	  p++;
+      }
+    return clean;
+}
+
 static int
 vdbf_create (sqlite3 * db, void *pAux, int argc, const char *const *argv,
 	     sqlite3_vtab ** ppVTab, char **pzErr)
@@ -115,6 +141,8 @@ vdbf_create (sqlite3 * db, void *pAux, int argc, const char *const *argv,
     char path[2048];
     char encoding[128];
     const char *pEncoding = NULL;
+    char ColnameCase[128];
+    const char *pColnameCase;
     int len;
     const char *pPath = NULL;
     gaiaDbfFieldPtr pFld;
@@ -123,13 +151,15 @@ vdbf_create (sqlite3 * db, void *pAux, int argc, const char *const *argv,
     int seed;
     int dup;
     int idup;
+    int text_dates = 0;
+    int colname_case = GAIA_DBF_COLNAME_LOWERCASE;
     char *xname;
     char **col_name = NULL;
     gaiaOutBuffer sql_statement;
     if (pAux)
 	pAux = pAux;		/* unused arg warning suppression */
 /* checking for DBF PATH */
-    if (argc == 5)
+    if (argc == 5 || argc == 6 || argc == 7)
       {
 	  pPath = argv[3];
 	  len = strlen (pPath);
@@ -156,12 +186,38 @@ vdbf_create (sqlite3 * db, void *pAux, int argc, const char *const *argv,
 	    }
 	  else
 	      strcpy (encoding, pEncoding);
+	  if (argc >= 6)
+	      text_dates = atoi (argv[5]);
+	  if (argc >= 7)
+	    {
+		pColnameCase = argv[6];
+		len = strlen (pColnameCase);
+		if ((*(pColnameCase + 0) == '\'' || *(pColnameCase + 0) == '"')
+		    && (*(pColnameCase + len - 1) == '\''
+			|| *(pColnameCase + len - 1) == '"'))
+		  {
+		      /* the charset-name is enclosed between quotes - we need to dequote it */
+		      strcpy (ColnameCase, pColnameCase + 1);
+		      len = strlen (ColnameCase);
+		      *(ColnameCase + len - 1) = '\0';
+		  }
+		else
+		    strcpy (ColnameCase, pColnameCase);
+		if (strcasecmp (ColnameCase, "uppercase") == 0
+		    || strcasecmp (ColnameCase, "upper") == 0)
+		    colname_case = GAIA_DBF_COLNAME_UPPERCASE;
+		else if (strcasecmp (ColnameCase, "samecase") == 0
+			 || strcasecmp (ColnameCase, "same") == 0)
+		    colname_case = GAIA_DBF_COLNAME_CASE_IGNORE;
+		else
+		    colname_case = GAIA_DBF_COLNAME_LOWERCASE;
+	    }
       }
     else
       {
 	  *pzErr =
 	      sqlite3_mprintf
-	      ("[VirtualDbf module] CREATE VIRTUAL: illegal arg list {dbf_path, encoding}");
+	      ("[VirtualDbf module] CREATE VIRTUAL: illegal arg list {dbf_path, encoding [ , text_dates [ , colname_case ]] }");
 	  return SQLITE_ERROR;
       }
     p_vt = (VirtualDbfPtr) sqlite3_malloc (sizeof (VirtualDbf));
@@ -172,6 +228,7 @@ vdbf_create (sqlite3 * db, void *pAux, int argc, const char *const *argv,
     p_vt->zErrMsg = NULL;
     p_vt->db = db;
     p_vt->dbf = gaiaAllocDbf ();
+    p_vt->text_dates = text_dates;
 /* trying to open file */
     gaiaOpenDbfRead (p_vt->dbf, path, encoding, "UTF-8");
     if (!(p_vt->dbf->Valid))
@@ -195,7 +252,10 @@ vdbf_create (sqlite3 * db, void *pAux, int argc, const char *const *argv,
 /* preparing the COLUMNs for this VIRTUAL TABLE */
     gaiaOutBufferInitialize (&sql_statement);
     xname = gaiaDoubleQuotedSql (argv[2]);
-    sql = sqlite3_mprintf ("CREATE TABLE \"%s\" (PKUID INTEGER", xname);
+    if (colname_case == GAIA_DBF_COLNAME_LOWERCASE)
+	sql = sqlite3_mprintf ("CREATE TABLE \"%s\" (pkuid INTEGER", xname);
+    else
+	sql = sqlite3_mprintf ("CREATE TABLE \"%s\" (PKUID INTEGER", xname);
     free (xname);
     gaiaAppendToOutBuffer (&sql_statement, sql);
     sqlite3_free (sql);
@@ -214,7 +274,9 @@ vdbf_create (sqlite3 * db, void *pAux, int argc, const char *const *argv,
     pFld = p_vt->dbf->Dbf->First;
     while (pFld)
       {
-	  xname = gaiaDoubleQuotedSql (pFld->Name);
+	  char *casename = convert_dbf_colname_case (pFld->Name, colname_case);
+	  xname = gaiaDoubleQuotedSql (casename);
+	  free (casename);
 	  dup = 0;
 	  for (idup = 0; idup < cnt; idup++)
 	    {
@@ -227,7 +289,9 @@ vdbf_create (sqlite3 * db, void *pAux, int argc, const char *const *argv,
 	    {
 		free (xname);
 		sql = sqlite3_mprintf ("COL_%d", seed++);
+		casename = convert_dbf_colname_case (sql, colname_case);
 		xname = gaiaDoubleQuotedSql (sql);
+		free (casename);
 		sqlite3_free (sql);
 	    }
 	  if (pFld->Type == 'N')
@@ -239,6 +303,15 @@ vdbf_create (sqlite3 * db, void *pAux, int argc, const char *const *argv,
 	    }
 	  else if (pFld->Type == 'F')
 	      sql = sqlite3_mprintf (", \"%s\" DOUBLE", xname);
+	  else if (pFld->Type == 'D')
+	    {
+		if (text_dates)
+		    sql =
+			sqlite3_mprintf (", \"%s\" VARCHAR(%d)", xname,
+					 pFld->Length);
+		else
+		    sql = sqlite3_mprintf (", \"%s\" DOUBLE", xname);
+	    }
 	  else
 	      sql =
 		  sqlite3_mprintf (", \"%s\" VARCHAR(%d)", xname, pFld->Length);
@@ -344,7 +417,9 @@ vdbf_read_row (VirtualDbfCursorPtr cursor, int *deleted_row)
 	  cursor->eof = 1;
 	  return;
       }
-    ret = gaiaReadDbfEntity (cursor->pVtab->dbf, cursor->current_row, &deleted);
+    ret =
+	gaiaReadDbfEntity_ex (cursor->pVtab->dbf, cursor->current_row, &deleted,
+			      cursor->pVtab->text_dates);
     if (!ret)
       {
 	  if (!(cursor->pVtab->dbf->LastError))	/* normal DBF EOF */
@@ -645,6 +720,17 @@ vdbf_eval_constraints (VirtualDbfCursorPtr cursor)
 					      if (ret >= 0)
 						  ok = 1;
 					      break;
+#ifdef HAVE_DECL_SQLITE_INDEX_CONSTRAINT_LIKE
+					  case SQLITE_INDEX_CONSTRAINT_LIKE:
+					      ret =
+						  sqlite3_strlike (pC->txtValue,
+								   pFld->
+								   Value->TxtValue,
+								   0);
+					      if (ret == 0)
+						  ok = 1;
+					      break;
+#endif
 					  };
 				    }
 				  break;
@@ -870,6 +956,17 @@ vdbf_rollback (sqlite3_vtab * pVTab)
     return SQLITE_OK;
 }
 
+static int
+vdbf_rename (sqlite3_vtab * pVTab, const char *zNew)
+{
+/* BEGIN TRANSACTION */
+    if (pVTab)
+	pVTab = pVTab;		/* unused arg warning suppression */
+    if (zNew)
+	zNew = zNew;		/* unused arg warning suppression */
+    return SQLITE_ERROR;
+}
+
 static int
 spliteVirtualDbfInit (sqlite3 * db)
 {
@@ -893,6 +990,7 @@ spliteVirtualDbfInit (sqlite3 * db)
     my_dbf_module.xCommit = &vdbf_commit;
     my_dbf_module.xRollback = &vdbf_rollback;
     my_dbf_module.xFindFunction = NULL;
+    my_dbf_module.xRename = &vdbf_rename;
     sqlite3_create_module_v2 (db, "VirtualDbf", &my_dbf_module, NULL, 0);
     return rc;
 }
